#ifndef SITEINFO_H
#define SITEINFO_H

#include <string>
#include <qstring.h>
#include <qprocess.h>
#include "globals.h"

class SiteInfo {
	public:
		SiteInfo();
		SiteInfo(const std::string &desc);
		~SiteInfo();
		void createCommand(QProcess *rsyncProcess, RSyncStuff::RSyncOperation op, QString passFilename = "");
		void setDescription(const std::string &d);
		std::string getDescription() const;
		void setHost(const std::string &h);
		std::string getHost() const;
		void setRemote(const std::string &r);
		std::string getRemote() const;
		void setLocalDir(const std::string &l);
		std::string getLocalDir() const;
		void setRemoteItem(const std::string &r);
		std::string getRemoteItem() const;
		void setLocalItem(const std::string &l);
		std::string getLocalItem() const;
		void setUser(const std::string &u);
		std::string getUser() const;
		void setPass(const std::string &p);
		std::string getPass() const;
		void setRootPass(const std::string &p);
		std::string getRootPass() const;
		void setComment(const std::string &c);
		std::string getComment() const;
		void setRsyncBin(const std::string &r);
		std::string getRsyncBin() const;
		void setPort(const int &p);
		void setPort(const std::string &p);
		std::string getPort() const;
		void setLocalhost(const bool &val);
		bool getLocalhost() const;
		void setAnonymous(const bool &val);
		bool getAnonymous() const;
		void setRootPerms(const bool &val);
		bool getRootPerms() const;
		void initInfo();
		void reset();
	private:
		QProcess *rsyncProcess;
		std::string description; // Human readable site description
		std::string host;				// Host for rsync source
		std::string remote;			// remote directory or rsync share
		std::string localDir;		// Local dir to browse under in gui
		std::string localItem;	// local file or some shit
		std::string remoteItem;	// remote file or some shit
		std::string user;				// Remote username
		std::string password;		// duh...
		std::string rootPass;		// root password, if applic. for full FS access
		std::string rsyncBin;		// actual rsync binary
		std::string comment;			// Misc info and shit
		int port;						// Remote rsync port
		bool modeDownload;	// Download or not (upload)?
		bool localhost;			// Both src/dest on localhost?
		bool anonymous;			// Anonymous user/no pass?
		bool useRoot;				// Start rsync as root for full FS access?
		int defaultPort;		// Just for convenience..
};

#endif /* SITEINFO_H */
