#include <iostream>
#include <assert.h>
#include <qapplication.h>
#include <qwidget.h>
#include "syncprocess.h"
#include <sys/stat.h>

using namespace std;

//SyncProcess::SyncProcess() : QObject(0, 0) {}

SyncProcess::SyncProcess(const SiteInfo &si, DlgStatus &ds, DlgShareList &dsl) :
		QObject(0, 0) {
	siteInfo = (SiteInfo *)&si;
	statusDialog = &ds;
	shareListDialog = &dsl;
	rsync = new QProcess((QString)siteInfo->getRsyncBin().c_str());
	ranOnce = false;
	loadingDir = true; // Initially we're listing shares.. only.. right?
	resetState();

	connect(rsync, SIGNAL(readyReadStdout()), this, SLOT(readOutput()));
	connect(rsync, SIGNAL(readyReadStderr()), this, SLOT(readError()));
	connect(rsync, SIGNAL(processExited()), this, SLOT(postSync()));
	connect(statusDialog, SIGNAL(syncViewDir(QString)), this, SLOT(viewDir(QString)));
	connect(statusDialog, SIGNAL(syncGetFile(QString)), this, SLOT(getFile(QString)));
	connect(statusDialog, SIGNAL(syncSendFile(QString)), this, SLOT(sendFile(QString)));

}

void SyncProcess::processOutput() {
	QString s = output[0];
	// skip 0 length lines or it stalls.  also count, if only listing shares.
	if (0 == s.length()) {
		if (noShareSpecified) {
			if (!s.isNull()) { // isNull, NOT isEmpty!
				motd += s; // otherwise is skipped over, not added to motd below
				motdCounter++;
			}
		}
		output.remove(output.first());
		return;
	}
	// try and match files here to skip all other matching...
	if (xferringFiles) {
		if ("generate_files finished" == s) {
			xferringFiles = false;
			cout << "Done transferring files." << endl;
		}
		// use leading space and % sign to match status lines.. good enough
		else if (s.startsWith(" ") && s.contains("%")) {
			// get rid of trailing carriage return
			if ('\r' == s[s.length() - 1])
				s.remove(s.length() - 1, 1);
			/* if there is still a CR, it's after one or more "lines" of data that it's
				 too late to process now.  dump them and process the most recent one. */
			if (s.contains('\r'))
				s.remove(0, s.findRev('\r') + 1);
			statusDialog->setFileProgress(atoi(s.section(' ', 0, 0,
				QString::SectionSkipEmpty)), s.section(' ', 1, 1,
				QString::SectionSkipEmpty).remove('%').toInt(), s.section(' ', 2, 2,
				QString::SectionSkipEmpty), s.section(' ', 3, 3, QString::SectionSkipEmpty));
		}
		// up to date files don't show a recv_files or whatever
		else if (s.endsWith(" is uptodate")) {
			currentFile = s.remove(s.findRev(' ', -2));
			statusDialog->setCurrFile(currentFile);
			statusDialog->makeFileComplete();
		}
		else if ("recv_files" == s.section('(', 0, 0)) {
			// fails with ()s in name: currentFile = s.section('(', 1).section(')', -2);
			// following 2 lines don't fail in such a case
			s.truncate(s.findRev(')'));
			currentFile = s.section('(', 1); 
			statusDialog->setCurrFile(currentFile);
		}
		else if (currentFile == s || "generate_files phase=1" == s
				|| "recv_files phase=1" == s || "generate_files phase=2" == s
				|| "recv_generator" == s.section('(', 0, 0)
				|| "renaming" == s.section(' ', 0, 0) || "got file_sum" == s
				|| s.endsWith("/")) {
			/* ignore these for now... directories and various other shit
				 that isn't useful atm (or ever).  not "unknown" stuff. */
		}
		else {
			QString temp = s.section(' ', 0, 2);
			if ("set modtime of" == temp)
				statusDialog->changeStatus("Setting modification times...");
			else if ("set gid of" == temp)
				statusDialog->changeStatus("Setting permissions...");
			else
				cout << "***Unknown fileline:" << endl << "\"" << s << "\"" << endl;
		}
	}
	else if (recvFileInfo) {
		if ("recv_file_list done" == s) {
			recvFileInfo = false;
		}
		else {
			// this # files to consider only shows up when not listing a module
			if (s.section(' ', 1) != "files to consider"
					&& s.section(' ', 1) != "file to consider")
				statusDialog->addListItem(s);
		}
	}
	else {
		// rsync outputs a full file list first.. collect this
		if (recvListStart) {
			// file list ends here
			if ("received" == s.section(' ', 0, 0)) {
				assert(atoi(s.section(' ', 1, 1)) == (fileCountList + fileRemovedList));
				recvListStart = false;
				recvListDone = true;
				recvFileInfo = true;
			}
			// until the above case, still collecting filenames
			else {
				int i = 0;
				QString name = s;
				if (s.section('(', 0, 0) != "recv_file_name") {
					// ...i think this became obsolete at some point? \r stuff?
					// try and remove stupid file counter prepended to output
					int index = s.find("recv_file_name(");
					cout << "Unexpected line:" << endl << "\"" << s << "\"" << endl;
				 	if (-1 != index)
						cout << ", fixing as: " << s.remove(0, index) << endl;
					else {
						cout << ", cannot fix." << endl;
						return;
					}
				}
				name.truncate(name.findRev(')'));
				name = name.section('(', 1); 
				if ((i = filenames.count()) > 0) {
					/* get rid of dirs.  if current filename includes last "filename" + /,
						 last one must be dir name instead */
					if (name.startsWith(filenames[i - 1] + "/")) {
						filenames.remove(filenames.last());
						fileCountList--;
						fileRemovedList++;
					}
				}
				filenames += name;
				fileCountList++;
			}
		}
		// this comes up after motd.. next is filename list
		// two cases, first for viewing rsync module, second for downloading
		else if (s.contains("[receiver] ") || s.contains("receiving file list ...")) {
			motdNext = false;
			motdReceived = true;
			recvListStart = true;
		}
		// everything here is the message of the day
		else if (motdNext && !motdReceived)
			motd += s;
		// this lets us know that the motd is coming up next
		else if (s.contains("opening tcp connection"))
			motdNext = true;
		else if (!listingOnly && "delta transmission enabled" == s) {
			xferringFiles = true;
			cout << "Transferring files." << endl;
			for (QStringList::iterator it = filenames.begin(); it != filenames.end(); it++)
				statusDialog->addTransferItem(*it);
		}
		else if (listingOnly) {
			QString t1 = s.section('=', 0, 0); // temp
			if ("generate_files phase" == t1 || "recv_files phase" == t1
					|| "generate_files finished" == s) {
				// ignore. not "unknown" output.
			}
		}
		else {
			QString t1 = s.section('=', 0, 0); // t = temp.  i hate wasting space.
			QString t2 = s.section('(', 0, 0); // t = temp.  i hate wasting space.
			QString t3 = s.section(' ', 0, 0); // t = temp.  i hate wasting space.
			if ("get_local_name count" == t1  || "recv_files" == t2
					|| "delta transmission enabled" == t1 || "generator starting pid" == t1
					|| "recv_files finished" == s || "sent" == t3 || "total" == t3
					|| "_exit_cleanup" == t2) {
				/* nothing to do with these...  maybe just ignore unknowns
					 later?  would rather catch REAL unknowns though.. (below) */
			}
			else
				cout << "***Unknown:" << endl << "\"" << s << "\"" << endl;
		}
	}
	output.remove(output.first());
	/* this is messy.  at this point, i don't care.  without it,
		 and with processEvents(), parsing stalls after rsync quits.
		 so, this works... until i figure out wtf i'm doing.
		 see similar note below */
	if (output.count() > 0)
		processOutput();
	else
		readOutput();
}

void SyncProcess::readOutput() {
	if (readingOutput)
		return;
	readingOutput = true;
	QString pOut;
	if (rsync->canReadLineStdout())
		output += rsync->readLineStdout();
	else
		output += QString(rsync->readStdout());
	readingOutput = false;
	processOutput();
	/* this is messy.  at this point, i don't care.  without it,
		 and with processEvents(), parsing stalls after rsync quits.
		 so, this works... until i figure out wtf i'm doing.
		 see similar note above */
	if (qApp->hasPendingEvents()) {
		qApp->processEvents();
	}
	if (rsync->canReadLineStdout())
		readOutput();
}

void SyncProcess::readError() {
	QString pError = rsync->readLineStderr();
	cerr << "___ERROR " << pError << endl;
}

void SyncProcess::go(RSyncStuff::RSyncOperation op) {
	if (ranOnce)
		resetState();

	// create temp password file with fairly random name
	srand(time(NULL));
	passFile.setName(((QString)"/tmp/qsync-%1").arg(rand()));
	
	// we need this to separate out MOTD from share listings
	if (!siteInfo->getRemote().length())
		noShareSpecified = true;

	siteInfo->createCommand(rsync, op, passFile.name());
	if (RSyncStuff::ChangeDir == op)
		listingOnly = true;
	else
		listingOnly = false;

	runProcess();
}

void SyncProcess::runProcess() {
	QValueList<ArgType>::Iterator it;
	writePassFile();
	filenames.clear();
	if (listingOnly)
		statusDialog->clearRecvList();
	// if we're not looking up the share list, bring up main file lists
	//if (statusDialog->isHidden() && !noShareSpecified) {
	if (!noShareSpecified) {
		statusDialog->setCaption(QString("%1::%2").arg(QString(siteInfo->getHost().c_str())).arg(QString(siteInfo->getRemote().c_str())));
		statusDialog->show();
	}
	cout << "Rsync args: " << rsync->arguments().join(" ") << endl;
	if (rsync->start()) {
		syncRunning = ranOnce = true;
		cout << "Rsync process started." << endl;
		syncStartTime = QDateTime::currentDateTime();
	}
	else
		cerr << "ERROR: failed to start rsync process." << endl;
}

void SyncProcess::viewDir(QString s) {
	siteInfo->setLocalDir("");
	syncOperation(RSyncStuff::ChangeDir, s);
}

void SyncProcess::getFile(QString s) {
	//siteInfo->setLocalDir((string)statusDialog->getCWD());
	syncOperation(RSyncStuff::DoDownload, s);
}

void SyncProcess::sendFile(QString s) {
	siteInfo->setLocalDir((string)statusDialog->getCWD());
	syncOperation(RSyncStuff::DoUpload, s);
}

void SyncProcess::syncOperation(RSyncStuff::RSyncOperation op, QString s) {
	QString temp;
	// set local dir from main dialog somewhere
	siteInfo->setLocalDir((string)statusDialog->getCWD());
	switch (op) {
		case RSyncStuff::ChangeDir:
			temp = siteInfo->getRemote().c_str();
			if (".." == s) {
		 		// check if was already a dir, if so strip up to 2nd slash
				if ('/' == temp[temp.length() - 1] && temp.contains('/') > 1)
					temp.truncate(temp.findRev("/", -2));
				else
					temp.truncate(temp.findRev("/"));
				temp += "/"; // re-add trailing slash or we don't see dir contents
			}
			else {
				temp += s + "/";
			}
			siteInfo->setRemote((string)temp);
			break;
		case RSyncStuff::DoDownload:
			/*
			temp = siteInfo->getRemote().c_str();
			if ('/' != temp[temp.length() - 1])
				temp.truncate(temp.findRev("/"));
			temp += s;
			siteInfo->setRemoteItem((string)temp);
			*/
			siteInfo->setRemoteItem((string)s);
			break;
		case RSyncStuff::DoUpload:
			/*
			temp = siteInfo->getLocalDir().c_str();
			if ('/' != temp[temp.length() - 1])
				temp.truncate(temp.findRev("/"));
			temp += s;
			siteInfo->setLocalItem((string)temp);
			*/
			siteInfo->setLocalItem((string)s);
			break;
		default:
		cerr << "ERROR: Bad syncOperation (" << __FILE__ << ": " << __LINE__ << ") -- You should never see this!" << endl;
		break;
	}

	go(op);
}

void SyncProcess::resetState() {
	readingOutput = readingError = readyOutput = readyError = motdReceived = connectDone
		= genRunning = recvDone = fileMatched
		= genFilesP1 = genFilesP2 = false;
	// new vars... above ones may go away
	recvFileInfo = motdNext = recvListStart = recvListDone = xferringFiles = false;
	modeIsDownload = listingOnly = true;
	fileCountList = fileRemovedList = filesRemain = generatorPID = motdCounter = 0;
	currentFile = "";
	motd.clear();
	filenames.clear();
	output.clear();
	shareListDialog->clearList();
	statusDialog->clearTransList();
}

void SyncProcess::postSync() {
	syncRunning = false;
	cout << "Rsync quit." << endl;
	passFile.remove();
	syncEndTime = QDateTime::currentDateTime();
	cout << "Elapsed time: " << syncStartTime.secsTo(syncEndTime) << " sec.\n" << endl;
	if (noShareSpecified) {
		processMOTDSpew();
		displayMOTD();
		shareListDialog->show();
		emit shareListShown();
		//TODO: is this needed anymore? testing only? displayShares();
	}
}

// processMOTDSpew() separates the combined MOTD and share list that rsync spews out together, blah
void SyncProcess::processMOTDSpew() {
	QStringList::Iterator it;
	QStringList motdTemp = motd;
	QString temp;
	motd.clear();
	for (it = motdTemp.begin(); it != motdTemp.end(); it++) {
		if (0 == motdCounter) { // following strings will be shares
			temp = (*it).remove('\t'); // the tabs are useless, mess up .section(...)
			if (!temp.section(" ", 1).remove(' '))
				shareListDialog->addShare(temp.section(" ", 0, 0), "(no description)");
			else
				shareListDialog->addShare(temp.section(" ", 0, 0),
					temp.section(" ", 1, 0xffffffff, QString::SectionSkipEmpty));
		}
		else { // still MOTD
			motd += *it;
			if ("" == *it) motdCounter--;
		}
	}
}

// displays the rsync server's message of the day.. but wait.. what if it doesn't exist?!?
void SyncProcess::displayMOTD() {
	/*
	for (QStringList::Iterator it = motd.begin(); it != motd.end(); it++)
		printStatus("_motd: " + *it);
		*/
	shareListDialog->setMotd(motd);
}

// display the rsync server's listed shares (doesn't show hidden ones, of course)
/*void SyncProcess::displayShares() {
	for (QStringList::Iterator it = shareList.begin(); it != shareList.end(); it++)
		cout << "_share: " << *it << endl;
} TODO: MAKE DlgShareList funtion that will print this shit.. if it's needed*/

void SyncProcess::setNoShareSpecified(bool b) {
	noShareSpecified = b;
}

void SyncProcess::writePassFile() {
	string temp;
	if (!passFile.open(IO_WriteOnly))
		cerr << "ERROR: Could not open password file for writing!" << endl;
	temp = siteInfo->getPass();
	for (unsigned int i = 0; i < temp.length(); i++) {
		if (passFile.putch(temp[i]) == -1)
			cerr << "ERROR: Could not write password to file!" << endl;
	}
	passFile.putch('\n');
	passFile.close();
	chmod(passFile.name(), S_IRUSR | S_IWUSR);
}

void SyncProcess::printStatus(QString s, bool isError, bool addEndl) {
	if (isError) cerr << s;
	else cout << s;
	//if (addEndl && isError) cout << endl;
	if (addEndl) cout << endl;
	else cout << flush;
}
